#!/usr/local/bin/perl
#
#   CUSTOM VERSION FOR SSL DOUBLE-PROXY; THIS IS THE SSL VERSION.
#   This should only run when the scheme is "https".
#   *REQUIRES* Perl 5, because of the Net::SSLeay module.
#
#   CGIProxy 1.2
#
#   nph-proxy.cgi-- CGI Proxy 1.2: an proxy in the form of a CGI script.
#     Retrieves the resource at any HTTP or FTP URL, updating embedded URLs
#     in HTML resources to point back through this script.  By default, no
#     user info is sent to the server.  Options include text-only proxying
#     to save bandwidth, cookie support, ad-filtering, and a user-defined
#     encoding of the target URL.  Runs in Perl 4 or 5.  
#
#   Copyright (C) 1996, 1998-1999 by James Marshall, james@jmarshall.com
#
#   For the latest, see http://www.jmarshall.com/tools/cgiproxy/
#
#
#   IMPORTANT NOTE ABOUT ANONYMOUS BROWSING:
#     Anonymity is mostly supported, but is not bulletproof.  For example, 
#     malicious JavaScript or embedded objects on pages can send your 
#     identity to the server.  CGIProxy was originally made for indirect
#     browsing more than anonymity, but since people are using it for 
#     anonymity, I'll make it as anonymous as possible.  Suggestions welcome.
#     For better anonymity, browse with JavaScript turned off.
#
#   CONFIGURATION:
#     None required.  Options in the "user configuration" section include:
#       . Set $TEXT_ONLY, $SUPPORT_COOKIES, $FILTER_ADS, and $REMOVE_SCRIPTS 
#           as desired.  Set $REMOVE_SCRIPTS if anonymity is important.
#       . To change the encoding format of the URL, modify the
#           &proxy_encode() and &proxy_decode() routines.  Make sure the
#           &proxy_encode() in startproxy.cgi matches the one here.  The
#           default routines are suitable for simple PATH_INFO compliance.
#       . If you're using another HTTP proxy, set $ENV{'http_proxy'} and
#           $ENV{'no_proxy'}.
#       . If filtering ads, you can customize this with a few settings.
#       . If supporting cookies, you can specify which cookies are allowed
#           and which are not.
#       . To make this run in Perl 4, follow the instructions around the 
#           "use Socket" line.
#       . Other minor config is possible; see the user configuration
#           section.
#       . If you're using Windows, you must comment out the line
#           "alarm(600);", because Windows doesn't seem to support alarms.
#
#     This script MUST be installed as a non-parsed header (NPH) script.
#       In Apache and many other servers, this is done by simply starting the
#       filename with "nph-".
#
#
#   TO USE:
#     Use the associated startproxy.cgi to start a session, or just call
#       this script with the encoded target URL in PATH_INFO.
#
#
#   I first wrote this in 1996 as an experiment to allow indirect browsing.
#     Parts of the code have been upgraded since then, like the addition of
#     HTML 4.0 tags.  The original seed was a program I wrote for Rich
#     Morin's article in the June 1996 issue of Unix Review, online at
#     http://www.cfcl.com/tin/P/9606.html.
#
#   Confession: I didn't originally write this with the spec for HTTP
#     proxies in mind, and there are probably some violations of the protocol
#     (at least for proxies).  This whole thing is one big violation of the
#     proxy model anyway, so I hereby rationalize that the spec can be widely
#     interpreted here.  If there is demand, I can make it more conformant.
#
#
#   TO DO:
#     since anonymity seems to be in demand, improve/clean up anonymity here
#     more error checking?
#     find a simple encryption technique for proxy_encode()
#     make able to be a real HTTP proxy, listening to a port (maybe make
#       modified version)
#     support more protocols, like SSL or gopher
#     support streaming media (probably requires HTTP 1.1)
#     Any (good) way to support HTTP Basic authentication?  Only way I can
#       think of uses an embedded form, cookies, and a special "admin:"
#       scheme-- messy.
#
#   BUGS:
#     Anonymity is not perfect.  In particular, JavaScript on a page can
#       compromise your anonymity if it wants to.
#     URLs generated by JavaScript or similar mechanisms won't be re-proxy'ed
#       correctly.  JavaScript in general may not work as expected.
#     Since ALL of your cookies are sent to this script (which then chooses
#       the relevant ones), some cookies could conceivably be dropped if
#       you accumulate a whole lot.  I haven't seen this happen yet.
#     The regex that converts URLs is not perfect.  A perfect regex would
#       be much slower.  See the notes above that block for more details.
#     When converted, URLs are not HTML-unescaped then re-escaped as they
#       should be.  To fix this would cost a lot of CPU time and would seldom
#       matter, but it's easy enough to do if there's demand.
#     Comments and script contents aren't parsed correctly, so sometimes
#       URLs within those will be mistakenly converted.  In reality, this
#       sometimes approximates what you want anyway.
#
#--------------------------------------------------------------------------
#
# Here are all global variables used in this script.  Some could be
#   rewritten as local variables, possibly with a performance loss.  With
#   these declared and all local() statements converted to my(), you can
#   (almost) compile with "use strict 'vars'", which is always a good idea
#   when programming in Perl 5.
#
# my($TEXT_ONLY, $SUPPORT_COOKIES, $FILTER_ADS, $SHOW_FTP_WELCOME,
#    $NO_COOKIE_WITH_IMAGE, @ALLOWED_COOKIE_SERVERS, @BANNED_COOKIE_SERVERS,
#    @BANNED_IMAGE_URL_PATTERNS, $RETURN_EMPTY_GIF,
#    $AF_INET, $SOCK_STREAM,
#    $PROXY_VERSION, $env_accept, $images_are_banned_here,
#    $URL, $scheme, $authority, $path, $host, $port,
#    $this_url, $base_url, $base_scheme, $base_host, $base_path,
#    $status, $headers, $body, @body, @welcome, $is_html,
#    ) ;
#
#--------------------------------------------------------------------------

#--------------------------------------------------------------------------
#    user configuration
#--------------------------------------------------------------------------

# Added to allow SSL redirection through another proxy.  Note trailing slash.
$this_url=     'http://.../' ;
$this_url_ssl= 'https://.../' ;


# If set, then proxy traffic will be restricted to text data only, to save 
#   bandwidth (though it can still be circumvented with uuencode, etc.).
$TEXT_ONLY= 0 ;      # set to 1 to allow only text data, 0 to allow all

# If set, then support cookies through the proxy.  See also
#   @ALLOWED_COOKIE_SERVERS and @BANNED_COOKIE_SERVERS below.
$SUPPORT_COOKIES= 1 ;

# If set, then adopt several tactics to minimize ads and ad-related cookies.
# If you want to fine-tune the behavior, configure $NO_COOKIE_WITH_IMAGE, 
#   @ALLOWED_COOKIE_SERVERS, @BANNED_COOKIE_SERVERS, and 
#   @BANNED_IMAGE_URL_PATTERNS below.
# To remove most popup advertisements, also set $REMOVE_SCRIPTS=1 below.
$FILTER_ADS= 0 ;


# If set, then <script>...</script> will be removed from HTML resources.
#   This will remove most, but not all, JavaScript and other scripts.
# It will NOT remove JavaScript contained in tag attributes, like
#   onMouseOver="...", etc., so a malicious server can still run any
#   JavaScript on your machine if you have JavaScript turned on.  SUCH
#   JAVASCRIPT CAN COMPROMISE YOUR ANONYMITY!  Just so you know.
# Note that turning this on removes most popup ads!  :)
$REMOVE_SCRIPTS= 0 ;


# Create your own proxy_encode() and proxy_decode() to tranform the target
#   URL to and from the format that will be stored in PATH_INFO.  The encoded
#   form should only contain characters that are legal in PATH_INFO.  This
#   varies by server, but using only printable chars, no "?" or "#", and no 
#   two adjacent slashes ("//") works on most servers.  Try not to make the
#   resulting string too long, either.
# Of course, proxy_decode() must exactly undo whatever proxy_encode() does.
# The proxy_encode() here must match the one in startproxy.cgi.
# Make proxy_encode() as fast as possible-- it's a major bottleneck for the 
#   whole program.

sub proxy_encode {
    local($URL)= @_ ;
    $URL=~ s#^([\w+.-]+)://#$1/# ;                 # http://xxx -> http/xxx
#    $URL=~ s/(.)/ sprintf('%02x',ord($1)) /ge ;   # each char -> 2-hex
#    $URL=~ tr/a-zA-Z/n-za-mN-ZA-M/ ;              # rot-13
    return $URL ;
}

sub proxy_decode {
    local($PATH_INFO)= @_ ;
#    $PATH_INFO=~ tr/a-zA-Z/n-za-mN-ZA-M/ ;        # rot-13
#    $PATH_INFO=~ s/([0-9A-Fa-f]{2})/ sprintf("%c",hex($1)) /ge ;
    $PATH_INFO=~ s#^([\w+.-]+)/#$1://# ;           # http/xxx -> http://xxx
    return $PATH_INFO ;
}



# set HTTP proxy if needed
#$ENV{'http_proxy'}= '' ;
#$ENV{'no_proxy'}= '' ;


# Various options to help filter ads and stop cookie-based privacy invasion.
# These are only effective if $FILTER_ADS is set above (except for
#   @ALLOWED_COOKIE_SERVERS).
# The default settings are OK, but the sample pattern lists are very simple
#   and could be expanded.  Improvements welcome!
# The champs of ad-filtering are JunkBusters, at http://www.junkbusters.com.
#   If ad-filtering is your primary motive, go there, dey da best.
# Note that the lists @ALLOWED_COOKIE_SERVERS and @BANNED_COOKIE_SERVERS
#   use Perl regular expressions (aka patterns or regexes), not literal
#   host names.  To turn a hostname into a pattern, replace every "." with 
#   "\.", add "^" to the beginning, and add "$" to the end.  For example,
#   "www.myhost.com" becomes "^www\.myhost\.com$".  To match *every* host
#   ending in something, leave out the "^".  For example, "\.myhost\.com$"
#   matches every host ending in ".myhost.com".
# @BANNED_IMAGE_URL_PATTERNS also uses Perl patterns.  If an image's URL
#   matches one of the patterns, it will not be downloaded (used for
#   ad-filtering).  For more information on Perl regular expressions, see
#   the Perl documentation.
# Note that most popup ads can be removed by setting $REMOVE_SCRIPTS=1, above.

$NO_COOKIE_WITH_IMAGE= 1 ;        # reject cookies attached to images
$RETURN_EMPTY_GIF= 1 ;            # replace banned images with 1x1 trans GIF

# If non-empty, only allow cookies from servers matching one of these patterns
# Comment this out to allow all cookies (subject to @BANNED_COOKIE_SERVERS).
@ALLOWED_COOKIE_SERVERS= ('\bslashdot\.org$') ;

# reject cookies from servers matching these patterns
@BANNED_COOKIE_SERVERS= (
    '\.doubleclick\.net$', 
    '\.preferences\.com$',
    '\.imgis\.com$',
    '\.adforce\.com$',
    '\.focalink\.com$',
    '\.flycast\.com$',
    '\.go\.com$',
    '\.avenuea\.com$',
    '\.linkexchange\.com$',
    '\.pathfinder\.com$',
    '\.burstnet\.com$',
    '\btripod\.com$',
    '\bgeocities\.yahoo\.com$',
    '\.mediaplex\.com$',
    ) ;

# reject images whose URL matches any of these patterns
@BANNED_IMAGE_URL_PATTERNS= (
    '\b[a-z]\.doubleclick\.net(:\d*)?/',
    '\.imgis\.com\b',
    '\.adforce\.com\b',
    '\.avenuea\.com\b',
    '\.go\.com(:\d*)?/ad/',
    '\.eimg\.com\b',
    '\bexcite\.netscape\.com(:\d*)?/.*/promo/',
    '/excitenetscapepromos/',
    '\.yimg\.com(:\d*)?.*/promo/',
    '\bus\.yimg\.com/[a-z]/(\w\w)/\1',
    '\bpromotions\.yahoo\.com(:\d*)?/promotions/',
    '\.cnn\.com(:\d*)?/ads/',
    'ads\.msn\.com\b',
    '\blinkexchange\.com\b',
    '\badknowledge\.com\b',
    '/SmartBanner/',
    '\bdeja\.com/ads/',
    '\bimage\.pathfinder\.com/sponsors',
    'ads\.tripod\.com',
    ) ;



# Set HTTP_USER_AGENT to something generic like this if you want to be extra
#   careful.  Conceivably, revealing which browser you're using may be a
#   slight privacy or security risk.
# However, note that some URLs serve different pages depending on which
#   browser you're using, so some pages will change if you set this.
# Leaving this commented out will just use the user's HTTP_USER_AGENT.
#$ENV{'HTTP_USER_AGENT'}= 'Mozilla/4.05 [en] (X11; I; Linux 2.0.34 i586)' ;


#------ end of (useful) user configuration --------------------------------

$PROXY_VERSION= '1.2' ;

# OK, let's time this thing
#($sutime,$sstime)= (times)[0,1] ;

# Fix submitted by Alex Freed:  Under some unidentified conditions,
#   instances of nph-proxy.cgi can hang around for many hours and drag the
#   system.  So until we figure out why that is, here's a 10-minute timeout.
#   Please write me with any insight into this, since I can't reproduce the
#   problem.  Under what conditions, on what systems, does it happen?
# 9-9-1999: One theory is that it's a bug in older Apaches, and is fixed by
#   upgrading to Apache 1.3.6 or better.  Julian Haight reports seeing the
#   same problem with other scripts on Apache 1.3.3, and it cleared up when
#   he upgraded to Apache 1.3.6.  Let me know if you can confirm this.
# Unfortunately, Windows doesn't seem to support alarms.  Comment out
#   the next two lines if you're using Windows.

$SIG{'ALRM'} = 'timeexit' ;
alarm(600);

# Exit upon timeout.  If you wish, add code to clean up and log an error.
sub timeexit { exit 1 }


# The CGI "standard" at NCSA is unclear on whether $ENV{'SCRIPT_NAME'} 
#   should have a leading slash, and different servers implement
#   it differently.  Here we just make it with a leading slash.
$ENV{'SCRIPT_NAME'}=~ s#^/?#/# ;


# Copy often-used environment vars into scalars, for efficiency
$env_accept= $ENV{'HTTP_ACCEPT'} || '*/*' ;     # may be modified later


# One user reported problems with binary files on certain other OS's, and
#   this seemed to fix it.  Supposedly, either this or the "binmode S"
#   statements below the newsocketto() calls work, or all; I'm putting all in.
#   Tell me anything new you figure out about this.
binmode STDOUT ;


# %urlsin was an associative array that listed tags and attributes that 
#   may contain URLs, declared like:
# %urlsin= ('a', 'href',
#           'applet', 'codebase',
#           'fig', 'src|imagemap',
#           'form', 'action|script',  ...
#          ) ;


#--------------------------------------------------------------------------
#    parse URL, make checks, and set various globals
#--------------------------------------------------------------------------

# Read the URL from PATH_INFO, stripping the leading slash

$URL= &proxy_decode(substr($ENV{'PATH_INFO'},1)) ;

# Set the query string correctly, from either $ENV{QUERY_STRING} or what's
#   already in $URL.
# The query string may exist in either the encoded URL or in the containing
#   URL, as $ENV{QUERY_STRING}.  If the former, then the query string was
#   (definitely?) in a referenced URL, while the latter most likely implies a
#   GET form input.  Either query string is valid, but form input takes
#   precedence-- if $ENV{QUERY_STRING} exists, it should be used over any
#   query string in the encoded URL.
# Note that Netscape does not pass any query string data that is part of the
#   URL in the <form action> attribute, which is probably correct behaviour.
#   For this program to act the same, it would need to strip the query 
#   string when updating all <form action> URLs, way below.
# Question:  Is there ever a valid case when both QUERY_STRINGs exist??

$URL=~ s/(\?.*)?$/?$ENV{'QUERY_STRING'}/ 
    if $ENV{'QUERY_STRING'} ne '' ;


# Parse the URL, using a regex modelled from the one in RFC 2396 (URI syntax),
#   appendix B.
# This assumes a hierarchical scheme; it won't work for e.g. mailto:
# "authority" is the combination of host, port, and possibly other info.
# Note that $path here will also contain any query component; it's more like
#   the request URI.
# Note that $URL is guaranteed to be an absolute URL with no "#" fragment, 
#   though this does little error-checking.  Note also that the old ";" 
#   parameters are now included in the path component.

($scheme, $authority, $path)= 
    ($URL=~ m#^([\w+.-]+)://([^/?]*)(.*)$#i) ;
$scheme=~ tr/A-Z/a-z/ ;
$path= "/$path" if $path!~ m#^/# ;   # if path is '' or contains only query

# Set $is_html if $path (minus query) ends in .htm or .html
$is_html= $path=~ /\.html?(\?|$)/i ;

# Alert the user to unsupported URL, with an intermediate page
&unsupported_warning($URL) unless ($scheme=~ /^(https)$/) ;

# Require a host to be present (for $base_url safety later)
# jsm-- actually, should restrict this to valid hostname characters.
&HTMLdie('The target URL cannot contain an empty host name.')
    unless $authority=~ /^\w/ ;


# Exclude non-text if it's not allowed.  Err on the side of allowing too much.
if ($TEXT_ONLY) {
    # First, forbid requests for filenames with non-text-type extensions
    # Extensions are roughly taken from Netscape's "Helper Preferences" screen
    #   (but that was in 1996).
    # Boy, I really need that list of binary file extensions.  Write me if
    #   you have one (August 1998).
    local($nontext)= 
	      'gif|jpeg|jpe|jpg|tiff|tif|png|bmp|xbm'   # images
            . '|mp2|mp3|wav|aif|aiff|au|snd'            # audios
            . '|avi|qt|mov|mpeg|mpg|mpe'                # videos
            . '|gz|Z|exe|gtar|tar|zip|sit|hqx|pdf'      # applications
            . '|ram|rm|ra' ;                            # others

    &nontextdie if ($path=~ /\.($nontext)(;|\?|$)/i) ;

    # Then, filter the "Accept:" header to accept only text
    $env_accept=~ s#\*/\*#text/*#g ;    # not strictly perfect
    $env_accept= join(', ', grep(m#^text/#i, split(/\s*,\s*/, $env_accept)) ) ;
    &nontextdie unless $env_accept ne '' ;
}


# Take steps to try removing banner ads
if ($FILTER_ADS) {

    # Determine now if images are banned from this URL (is used later also)
    foreach (@BANNED_IMAGE_URL_PATTERNS) {
	$images_are_banned_here= 1, last if $URL=~ /$_/ ;
    }

    # Intercept request if it looks like an image is requested, i.e. unless
    #   the Accept: header allows either text/... or */... .
    if ($images_are_banned_here) {
	&skip_image unless 
	    grep(m#^(text|\*)/#i, split(/\s*,\s*/, $env_accept) ) ;
    }
}


$*= 1 ;     # allow multi-line matching


# Full elapsed time, not just CPU time
# $starttime= time ;


# Calculate $this_url for use later in &full_url().  It's an integral part
#   of &full_url(), placed here for speed, similar to the variables set in
#   &fix_base_vars.
# $this_url is an absolute URL pointing to this proxy script, complete with
#   trailing slash.  It has to be absolute to serve those HTML attributes
#   that require absolute URLs, like <base href>.
# Only include SERVER_PORT if it's not 80

# Changed-- now it's set at top, for double-SSL proxy
#$this_url= join('', 'http://', $ENV{'SERVER_NAME'}, 
#		   ($ENV{'SERVER_PORT'}==80  ? ''  : ':'.$ENV{'SERVER_PORT'} ),
#                    $ENV{'SCRIPT_NAME'}, '/') ;


# $base_url must be set correctly at any time &full_url() may be called.
#   &fix_base_vars() must be called as well, to set $base_scheme, $base_host 
#   and $base_path.
# Unfortunately, the base URL may change over the course of this program.  We
#   will keep it set based on whatever info we have so far, i.e. request URI,
#   then e.g. HTTP response headers, then e.g. <base> tag (which happens to
#   be in the reverse order of the ultimate precedence).
$base_url= $URL ;
&fix_base_vars ;   # must be called whenever $base_url is set


#--------------------------------------------------------------------------
#    Retrieve the resource into $body or @body using the correct scheme, 
#      also setting $status, $headers, and $is_html (all globals).
#      $is_html indicates whether the original resource is HTML, not
#      if a generated response is in HTML (e.g. an error message).
#--------------------------------------------------------------------------

if ($scheme eq 'https') {
    &https_get ;
}


#--------------------------------------------------------------------------
#    Modify entire response to point back through this script
#--------------------------------------------------------------------------

# NOTE: IT IS IMPORTANT TO DO THIS AS COMPLETELY AS POSSIBLE.  IF A
# USER UNKNOWINGLY GOES TO PAGE DIRECTLY AND NOT THROUGH THIS PROXY,
# HE/SHE MAY REVEAL HIM/HERSELF IN AN UNINTENDED WAY.

#--------------------------------------------------------------------------
# These were notes to myself from testing the speed of different methods.
#   Names like "nph-proxy2" refer to different modifications.  This version,
#   the fastest, was called nph-proxy2b.
#
# If YOU figure out a faster method, please tell me about it!
#
#--------------------------------------------------------------------------
#
# [This version is nph-proxy2b:  Break into @body array, do not use 
#   %urlsin to test if tags should be updated, replacement strings 
#   use (a|b|c) syntax.]
#
# This is by far the most time-consuming part of the script, the updating
#   of all URLs in an HTML file.
#
# Results of informal speed testing (not what I expected):
#   Breaking into @body array instead of one big string definitely saves 
#       significant time-- compare nph-proxy2 to nph-proxy1.  One test
#       showed a time saving of 1/3 to 1/2.
#   Using a %urlsin array does NOT seem to save time, even when only used
#       as boolean test to see if tag might contain URL-- compare nph-proxy2
#       to nph-proxy3, nph-proxy4, and (most similar) nph-proxy5.  Oh well.
#   It seems that reading one tag at a time, converting, and sending it
#       through does NOT save elapsed time over reading all tags before
#       converting, like I thought it would.  Both the CPU and elapsed
#       time are longer for one-tag-at-a-time approach-- compare
#       nph-proxy2 and nph-piper2.
#
#   Results of nph-proxy2 (blocks for each tag) to nph-proxy2b 
#   (single "(att1|att2|att3)" style regex) testing:
#       Mixed results, but overall, using a single regex (nph-proxy2b)
#       takes less CPU "user time", about the same "system time", and
#       slightly more elapsed time than nph-proxy2; I don't know why.
#       The elapsed time is slightly more both within the script and at
#       the HTTP client's end.  All differences are less than 10% on
#       average, and nph-proxy2b occasionally shows LESS elapsed time
#       than nph-proxy2.  CPU time is always less for nph-proxy2b than
#       for nph-proxy2.
#   Since the bottleneck of CPU time is tighter than for elapsed time,
#       and since the elapsed-time loss is less than the CPU time gain,
#       let's go with nph-proxy2b (not that it really makes much difference).
#       Besides, the code is easier to read that way.
#
# So basically, breaking into @body array helps, but not much else does.
#
#--------------------------------------------------------------------------
# 8-4-98 JSM: Found a bug in the regex, so changed it.  It more resembles
#   nph-proxy2.cgi now.
#--------------------------------------------------------------------------


# If the resource is HTML (and not empty), update all URLs in all tags that 
#   refer to URLs.

if ( $is_html  && ($body[0] ne '') ) {

    # Set $base_url if there is an HTML <base> tag.
    foreach (@body) {
	if (/<\s*base\b/i) {
	    $base_url= &HTMLunescape($1), &fix_base_vars, last
		if m#<\s*base\b[^>]*\bhref\s*=\s*"?([\w+.-]+://[^\s">]+)#i ;
	}
	last if m#<\s*(/head|body)\b#i ; # stop looking after </head> or <body>
    }

    # Strip everything between <script> and </script> if so configured.
    if ($REMOVE_SCRIPTS) {
	local($s) ;  # sequence number in .. test
	foreach (@body) { 
	    # Note the use of the seldom-seen ".." scalar operator.
	    if ( $s=(/<\s*script\b/i .. /<\s*\/script\b/i) ) {
		# Don't remove the part before the initial "<script>".
		if ($s==1) {
		    s/<\s*script\b.*>//i ;
		} else { 
		    $_= '' ;
		}
	    }
	}
    }

    # This is a complete list of HTML tags/attributes that may include a URL,
    #   compiled from the list at Sandia (read on 9-17-96, but last updated on
    #   12-7-95), and the latest HTML DTD as of 9-17-96.  The Sandia list
    #   includes non-standard tags found to be used by Netscape or Microsoft.
    # List updated 7-31-98 by JSM to include all URL-type attributes defined
    #   in HTML 4.0.
    # If anyone knows of a well-maintained list of standard and non-standard
    #   tags/attributes with URLs in them, please let me know!!
    # Note that most of these are very rarely used, if ever.  They're included
    #   for safety, since we don't want an anonymous user accidentally
    #   revealing themselves because of a non-anonymized URL.
    # This regex favors speed over readability.
    # This could probably be more efficient, but the solution eludes me.
    # Note that this is not a perfect regex.  The substitutions can be fooled
    #   by, e.g., <input value="<a href=bad"> or <input value="src=bad">.
    #   Also, it doesn't allow single quotes in URLs, which are legal,
    #   though rare.  It could be fixed with a slower regex.
    # The regex terminates URLs at spaces; URLs with spaces are disallowed,
    #   but fragments sometimes contain them anyway.
    # Denoting which of these are for images/binaries might be helpful, if
    #   we need more elaborate text-only support.
    # 8-4-98 JSM:  Changed the regex to fix a bug: would only change the 
    #   first URL in a tag.  New regex is probably slower.

    foreach (@body) {

	# Put the most common cases first

        s/(<[^>]*\bhref\s*=\s*["']?)([^\s"'>]*)/       $1 . &full_url($2) /ie,
            next if /<\s*a\b/i ;

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
        s/(<[^>]*\blowsrc\s*=\s*["']?)([^\s"'>]*)/     $1 . &full_url($2) /ie,
        s/(<[^>]*\blongdesc\s*=\s*["']?)([^\s"'>]*)/   $1 . &full_url($2) /ie,
        s/(<[^>]*\busemap\s*=\s*["']?)([^\s"'>]*)/     $1 . &full_url($2) /ie,
        s/(<[^>]*\bdynsrc\s*=\s*["']?)([^\s"'>]*)/     $1 . &full_url($2) /ie,
            next if /<\s*img\b/i ;

        s/(<[^>]*\bbackground\s*=\s*["']?)([^\s"'>]*)/ $1 . &full_url($2) /ie,
            next if /<\s*body\b/i ;

        s/(<[^>]*\bhref\s*=\s*["']?)([^\s"'>]*)/       $1 . &full_url($2) /ie,
            next if /<\s*base\b/i ;     # has special significance

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
        s/(<[^>]*\blongdesc\s*=\s*["']?)([^\s"'>]*)/   $1 . &full_url($2) /ie,
            next if /<\s*frame\b/i ;

        s/(<[^>]*\baction\s*=\s*["']?)([^\s"'>]*)/     $1 . &full_url($2) /ie,
        s/(<[^>]*\bscript\s*=\s*["']?)([^\s"'>]*)/     $1 . &full_url($2) /ie,
            next if /<\s*form\b/i ;     # needs special attention

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
        s/(<[^>]*\busemap\s*=\s*["']?)([^\s"'>]*)/     $1 . &full_url($2) /ie,
            next if /<\s*input\b/i ;

        s/(<[^>]*\bhref\s*=\s*["']?)([^\s"'>]*)/       $1 . &full_url($2) /ie,
            next if /<\s*area\b/i ;

        s/(<[^>]*\bcodebase\s*=\s*["']?)([^\s"'>]*)/   $1 . &full_url($2) /ie,
        s/(<[^>]*\bcode\s*=\s*["']?)([^\s"'>]*)/       $1 . &full_url($2) /ie,
        s/(<[^>]*\bobject\s*=\s*["']?)([^\s"'>]*)/     $1 . &full_url($2) /ie,
        s/(<[^>]*\barchive\s*=\s*["']?)([^\s"'>]*)/    $1 . &full_url($2) /ie,
            next if /<\s*applet\b/i ;


        # These are seldom-used tags, or tags that seldom have URLs in them

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
            next if /<\s*bgsound\b/i ;  # Microsoft only

        s/(<[^>]*\bcite\s*=\s*["']?)([^\s"'>]*)/       $1 . &full_url($2) /ie,
            next if /<\s*blockquote\b/i ;

        s/(<[^>]*\bcite\s*=\s*["']?)([^\s"'>]*)/       $1 . &full_url($2) /ie,
            next if /<\s*del\b/i ;

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
            next if /<\s*embed\b/i ;    # Netscape only

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
        s/(<[^>]*\bimagemap\s*=\s*["']?)([^\s"'>]*)/   $1 . &full_url($2) /ie,
            next if /<\s*fig\b/i ;      # HTML 3.0

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
            next if /<\s*h[1-6]\b/i ;   # HTML 3.0

        s/(<[^>]*\bprofile\s*=\s*["']?)([^\s"'>]*)/    $1 . &full_url($2) /ie,
            next if /<\s*head\b/i ;

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
            next if /<\s*hr\b/i ;       # HTML 3.0

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
        s/(<[^>]*\blongdesc\s*=\s*["']?)([^\s"'>]*)/   $1 . &full_url($2) /ie,
            next if /<\s*iframe\b/i ;

        s/(<[^>]*\bcite\s*=\s*["']?)([^\s"'>]*)/       $1 . &full_url($2) /ie,
            next if /<\s*ins\b/i ;

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
            next if /<\s*layer\b/i ;

        s/(<[^>]*\bhref\s*=\s*["']?)([^\s"'>]*)/       $1 . &full_url($2) /ie,
        s/(<[^>]*\burn\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
            next if /<\s*link\b/i ;

        s/(<[^>]*\burl\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
            next if /<\s*meta\b/i ;     # Netscape only

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
            next if /<\s*note\b/i ;     # HTML 3.0

        s/(<[^>]*\busemap\s*=\s*["']?)([^\s"'>]*)/     $1 . &full_url($2) /ie,
        s/(<[^>]*\bcodebase\s*=\s*["']?)([^\s"'>]*)/   $1 . &full_url($2) /ie,
        s/(<[^>]*\bdata\s*=\s*["']?)([^\s"'>]*)/       $1 . &full_url($2) /ie,
        s/(<[^>]*\barchive\s*=\s*["']?)([^\s"'>]*)/    $1 . &full_url($2) /ie,
        s/(<[^>]*\bclassid\s*=\s*["']?)([^\s"'>]*)/    $1 . &full_url($2) /ie,
        s/(<[^>]*\bname\s*=\s*["']?)([^\s"'>]*)/       $1 . &full_url($2) /ie,
            next if /<\s*object\b/i ;

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
        s/(<[^>]*\bimagemap\s*=\s*["']?)([^\s"'>]*)/   $1 . &full_url($2) /ie,
            next if /<\s*overlay\b/i ;  # HTML 3.0

        s/(<[^>]*\bcite\s*=\s*["']?)([^\s"'>]*)/       $1 . &full_url($2) /ie,
            next if /<\s*q\b/i ;

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
        s/(<[^>]*\bfor\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
            next if /<\s*script\b/i ;

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
            next if /<\s*select\b/i ;   # HTML 3.0

        s/(<[^>]*\bsrc\s*=\s*["']?)([^\s"'>]*)/        $1 . &full_url($2) /ie,
            next if /<\s*ul\b/i ;       # HTML 3.0

        s/(<[^>]*\bbackground\s*=\s*["']?)([^\s"'>]*)/ $1 . &full_url($2) /ie,
            next if /<\s*td\b/i ;       # Netscape extension?

        s/(<[^>]*\bbackground\s*=\s*["']?)([^\s"'>]*)/ $1 . &full_url($2) /ie,
            next if /<\s*th\b/i ;       # Netscape extension?

        s/(<[^>]*\bbackground\s*=\s*["']?)([^\s"'>]*)/ $1 . &full_url($2) /ie,
            next if /<\s*tr\b/i ;       # Netscape extension?

        s/(<[^>]*\bbackground\s*=\s*["']?)([^\s"'>]*)/ $1 . &full_url($2) /ie,
            next if /<\s*table\b/i ;    # Netscape extension?

    }   # foreach (@body)


    # Does this ever cause a problem, putting a comment before the entire body?
    unshift(@body, "<!-- resource has been modified by proxy script -->\n" ) ;

    # Change Content-Length header, since we changed the content
    $headers=~ s/^Content-Length:.*\012/ 
                 'Content-Length: ' . length(join('',@body)) . "\015\012"/ie ;

}


#--------------------------------------------------------------------------
#    Send response back to user
#--------------------------------------------------------------------------

# print the status line, headers, and the entire (possibly modified) resource
if ($ENV{'REQUEST_METHOD'} eq 'HEAD') {
    print $status, $headers ;
} elsif ($is_html) {
    print $status, $headers, @body ;
} else {
    print $status, $headers, $body ;
}


# Put this back in to run speed trials
#if (($headers=~ m#^Content-type:\s*text/html#i) || !$headers) {
#    # OK, let's time this thing
#    ($eutime,$estime)= (times)[0,1] ;
#    open(LOG,">>proxy2b.log") ;
#    print LOG "full times: ", $eutime-$sutime, " ", $estime-$sstime, 
#        " ", time-$starttime, "  URL: $URL\n" ;
#    close(LOG) ;
#}


exit ;

#--------------------------------------------------------------------------
#   DONE!!
#--------------------------------------------------------------------------


# Returns the full absolute URL to query our script for the given URI
#   reference.  PATH_INFO will include the encoded absolute URL of the target,
#   but the fragment will be appended unencoded so browsers will resolve it
#   correctly.
# This is a major bottleneck for the whole program, so speed is important here.
# Note that the calculations of $this_url, $base_scheme, $base_host, and
#   $base_path throughout the program are an integral part of this routine,
#   placed elsewhere for speed.
# This was rewritten 2-18-99 to be much more flexible, e.g. to allow encoding
#   and avoid illegal PATH_INFO sequences.  Unfortunately, it is a little 
#   slower.
# For HTTP, The URL to be encoded should include everything that is sent in
#   the request, including any query, but not any fragment.
# This only returns absolute URLs, though relative URLs would usually suffice.
#   If it matters, we could have a fullrelurl() and fullabsurl(), the latter
#   used for those HTML attributes that require an absolute URL (like <base>).
#
# The ?:?:?: statement resolves relative URLs to absolute URLs, given the
#   $base_{url,scheme,host,path} variables figured earlier.  It does it
#   simply and efficiently, and accurately enough; the full procedure is
#   described in RFC 2396 (URI syntax), section 5.2.
# RFC 2396, section 5 states that there are three types of relative URIs:
#   net_path (beginning with //, rarely used), abs_path (beginning with /),
#   and rel_path, any of which may be followed by a "?query"; the query must
#   be included in the result.  Thus, we only need to examine the start of
#   the relative URL.
# This ?:?:?: statement passes all test cases in RFC 2396 appendix C, except
#   for the following:  It does not reduce . and .. path segments (to do
#   so would take a lot more time), and it assumes $uri_ref has something
#   other than an empty fragment in it, i.e. that the URI is non-empty.
# This only works for hierarchical schemes, like HTTP or FTP.  Conceivably,
#   there's a problem if the base URL uses a non-hierarchical scheme, and
#   the document contains relative URLs.  Absolute URLs will be OK.
# Any HTML-escaping/unescaping should be done outside of this routine, since
#   it is used for any relative->absolute URL conversion, not just HTML.

sub full_url {
    local($uri_ref)= @_ ;

    # Separate fragment from URI
    local($uri,$frag)= $uri_ref=~ /^([^#]*)(#.*)?/ ;
    return $uri_ref if $uri eq '' ;  # allow bare fragments to pass unchanged

    # calculate absolute URL based on three possible cases
    local($absurl)= 
            $uri=~ m#^[\w+.-]*:#i   ?  $uri                 # absolute URL
          : $uri=~ m#^//#           ?  $base_scheme . $uri  # net_path (rare)
          : $uri=~ m#^/#            ?  $base_host . $uri    # abs_path, rel URL
          :                            $base_path . $uri ;  # relative path

    # Changed to accommodate SSL double-proxy
    return $absurl=~ m#^https://#i
	? join('', $this_url_ssl, &proxy_encode($absurl), $frag )
	: join('', $this_url,     &proxy_encode($absurl), $frag ) ;
}


# Set globals $base_url, $base_scheme, $base_host, and $base_path, based on 
#   value of $base_url.  This must be called whenever $base_url is set, which
#   unfortunately may vary over the course of the program.
# These are an integral part of &full_url(), placed outside of that for speed.
# To specify:
#   $base_scheme is the scheme of the base URL, ending in ":", like "http:".
#   $base_host is the scheme/host/port of the base URL, with no final slash.
#   $base_path is the scheme/host/port/path, through final slash.
# These are only relevant (and accurate) for hierarchical "/"-using schemes,
#   like HTTP or FTP.
# Any HTML-escaping/unescaping should be done outside of this routine.
sub fix_base_vars {
    # Guarantee that $base_url has at least a path of '/', inserting before 
    #   ?query if needed.
    $base_url=~ s#^([\w+.-]+://[^/?]+)/?#$1/# ;

    ($base_scheme)= $base_url=~ m#^([\w+.-]+:)//# ;
    ($base_host)=   $base_url=~ m#^([\w+.-]+://[^/?]+)# ; # no ending slash
    ($base_path)=   $base_url=~ m#^([^?]*/)# ;            # use greedy matching
}



#--------------------------------------------------------------------------
#    Scheme-specific routines
#--------------------------------------------------------------------------

#
# <scheme>_get: get resource at URL and set globals $status, $headers, 
#   $is_html, and either $body or @body depending on $is_html.  They are
#   globals for speed, to prevent unneeded copying of huge strings.
#

# https_get: actually supports both GET and POST

sub https_get {
    local($portst, $realhost, $realport, $request_uri, $request, $reply,
          $cookie_to_server, $lefttoget, $postbody) ;
    local($/)= "\012" ;

    use Net::SSLeay qw(sslcat) ;

    ($host,$port)= $authority=~ /^([^:]*):?(.*)$/ ;
    $host=~ tr/A-Z/a-z/ ;
    $port= 443 if $port eq '' ;

    # Some servers don't like default port in a Host: header, so use $portst.
    $portst= ($port==443)  ? ''  : ":$port" ;

    $realhost= $host ;
    $realport= $port ;
    $request_uri= $path ;

    # there must be a smoother way to handle proxies....
    if ($ENV{'http_proxy'}) {
	local($dont_proxy) ;
	foreach (split(/\s*,\s*/, $ENV{'no_proxy'})) {
	    $dont_proxy= 1, last if $host=~ /$_$/ ;
	}
	unless ($dont_proxy) {
	    (undef,$realhost,$realport)=
		$ENV{'http_proxy'}=~ m#^(http://)?([^/?:]*):?([^/?]*)#i ;
	    $realport= 443 if $realport eq '' ;
	    $request_uri= $URL ;
	}
    }


    # Changed to support SSL: First create $request, then use sslcat()

    #------ Create $request, use sslcat(), and get $reply -----------

    # Accept: and User-Agent: headers are needed because they affect results.
    # We're not fully supporting HTTP 1.1, but at least send the Host: header.
    # We're anonymously browsing, so don't include From: or Referer: headers.
    #   The User-Agent: header is a very slight privacy risk, but some pages
    #   load differently with different browsers.

    $request= join('', 
	    $ENV{'REQUEST_METHOD'}, ' ', $request_uri, " HTTP/1.0\015\012",
	    'Host: ', $host, $portst, "\015\012",    # being a good netizen
            'Accept: ', $env_accept, "\015\012",        # possibly modified
            'User-Agent: ', $ENV{'HTTP_USER_AGENT'}, "\015\012"
		   ) ;

    # Add the cookie if it exists and cookies are supported.
    if ($SUPPORT_COOKIES && ($ENV{'HTTP_COOKIE'} ne '')) {
	$cookie_to_server= 
	    &cookie_to_server($ENV{'HTTP_COOKIE'}, $path, $host) ;
	$request.= join('', 'Cookie: ', $cookie_to_server, "\015\012")
	    if $cookie_to_server ne '' ;
    }

    # If request method is POST, copy content headers and body to request.
    #   Loop to guarantee all is read from STDIN.
    if ($ENV{'REQUEST_METHOD'} eq 'POST') {
	$lefttoget= $ENV{'CONTENT_LENGTH'} ;
	$request.= join('', 
		'Content-type: ', $ENV{'CONTENT_TYPE'}, "\015\012", 
                'Content-length: ', $lefttoget, "\015\012\015\012"
			) ;
	do {
	    $lefttoget-= read(STDIN, $postbody, $lefttoget) ;
	    $request.= $postbody ;
	} while $lefttoget && length($postbody) ;

    # For GET requests, just add extra blank line
    } else {
	$request.= "\015\012" ;
    }


    # Call the all-purpose sslcat() routine from Net::SSLeay.
    # Kind of coarse, but we'll address problems as we see them.

    $reply= Net::SSLeay::sslcat($realhost, $realport, $request) ;


    #------ Copy full response into $status, $headers, and $body/@body ----

    # HTTP 1.x
    if ($reply=~ m#^HTTP/#) {
	($status, $headers, $body)=
	    $reply=~ /^(.*?\015?\012)(.*?\015?\012\015?\012)(.*)$/s ;

	# Unfold long header lines, a la RFC 822 section 3.1.1
	$headers=~ s/(\015\012|\012)[ \t]/ /g ;

	# If we're text only, then cut off non-text responses (but allow 
	#   unspecified types).
	if ($TEXT_ONLY) {
	    if ($headers=~ m#^Content-type:\s*([\w/]*)#i) {
		&nontextdie unless $1=~ m#^text/#i ;
	    }
	}

	# If server is one of @BANNED_IMAGE_URL_PATTERNS, then skip the image
	#   unless it's clearly a text type.
	if ($images_are_banned_here) {
	    &skip_image
		unless ($headers=~ m#^Content-type:\s*text/#i) ;
	}

	# Keeping $base_url, $base_host, and $base_path up-to-date is an 
	#   ongoing job.  Here, we look in appropriate headers.  Note that if
	#   Content-Base: doesn't exist, Content-Location: is an absolute URL.
	if        ($headers=~ m#^Content-Base:\s*([\w+.-]+://\S+)#i) {
	    $base_url= $1, &fix_base_vars ;
	} elsif   ($headers=~ m#^Content-Location:\s*([\w+.-]+://\S+)#i) {
	    $base_url= $1, &fix_base_vars ;
	} elsif   ($headers=~ m#^Location:\s*([\w+.-]+://\S+)#i) {
	    $base_url= $1, &fix_base_vars ;
	}

	# Now, fix the headers with &http_fix().  It uses &full_url(), and
	#   may modify the headers we just extracted the base URL from.
	# This also includes cookie support.
        &http_fix ;


	# Set $is_html if headers indicate HTML response
	$is_html= $is_html || ($headers=~ m#^Content-type:\s*text/html\b#i) ;

        # Only read body if the request method is not HEAD
        if ($ENV{'REQUEST_METHOD'} ne 'HEAD') {
	    if ($is_html) {
		@body= $body=~ /([^>]*>?)/g ;
	    }
	} else {
	    $body= ''; @body= () ;
	}


    # HTTP 0.9
    # Ignore possibility of HEAD, since it's not defined in HTTP 0.9
    } else {
	$is_html= 1 ;   # HTTP 0.9 by definition implies an HTML response

	# split $body into @body; with each element ending in ">" or EOstring
	@body= $reply=~ /([^>]*>?)/g ;

    }


}  # sub https_get()



#--------------------------------------------------------------------------

#
# <scheme>_fix: modify response as appropriate for given protocol (scheme).
#

# http_fix: modify headers as needed, including cookie support.
sub http_fix {

    # First, modify any appropriate HTTP headers to point back through the
    #   proxy.  Location: is most important.
    # Note that all these are absolute URIs, except possibly Content-Location:
    #   or URI:, which may be relative to Content-Base or the request URI--
    #   notably, NOT relative to anything in the content, like a <base> tag.
    $headers=~ s/^Content-Base:\s*(\S*)/ 'Content-Base: ' . &full_url($1)/gie ;
    $headers=~ s/^Content-Location:\s*(\S*)/
	                             'Content-Location: ' . &full_url($1)/gie ;
    $headers=~ s/^Location:\s*(\S*)/     'Location: '     . &full_url($1)/gie ;
    $headers=~ s/^URI:\s*(\S*)/          'URI: '          . &full_url($1)/gie ;

    # And the non-standard Refresh: header... any others?
    $headers=~ s/^(Refresh:.*;\s*URL=)(\S*)/ $1           . &full_url($2)/gie ;


    # Modify cookies to point back through the script, or they won't work.
    # If they're not supported, then filter them all out.
    if ($SUPPORT_COOKIES) {
	local($done) ;    # indicates cookie headers have been stripped

	# If @ALLOWED_COOKIE_SERVERS exists, then restrict to only those
	if (@ALLOWED_COOKIE_SERVERS) {
	    local($ok) ;
	    foreach (@ALLOWED_COOKIE_SERVERS) {
		$ok= 1, last   if $host=~ /$_/ ;
	    }
	    $done= 1, $headers=~ s/^Set-Cookie:.*\012//gi   unless $ok ;
	}

	# Take appropriate action if we're trying to filter ads out
	if ($FILTER_ADS) {

	    # If so configured, filter out cookies sent with non-text resources
	    # This actually covers all non-text resources, not just images.
	    if (!$done && $NO_COOKIE_WITH_IMAGE) {
		$done= 1, $headers=~ s/^Set-Cookie:.*\012//gi
		    unless $headers=~ m#^Content-Type:\s*text/#i ;
	    }

	    # Strip the cookies if source matches any of @BANNED_COOKIE_SERVERS
	    unless ($done) {
		foreach (@BANNED_COOKIE_SERVERS) {
		    $headers=~ s/^Set-Cookie:.*\012//gi, last if $host=~ /$_/ ;
		}
	    }
	}

	$headers=~ s/^Set-Cookie:\s*([^\015\012]*)/
	             'Set-Cookie: ' . &cookie_to_client($1, $path, $host)
		    /gie
	    unless $done ;

    } else {
	$headers=~ s/^Set-Cookie:.*\012//gi ;
    }

}


#--------------------------------------------------------------------------
#    Cookie routines
#--------------------------------------------------------------------------

# Cookie support:  The trick is how to send a cookie back to the client that
#   it will return for appropriate hosts.  Given that the target URL may be
#   encoded, and the client can't always tell where the target URL is, the 
#   only way to do that is to get *all* the cookies from the client and send
#   along the matching ones.  If the client has a lot of cookies through the
#   proxy, this could conceivably be a problem.  Oh well, it works for the
#   limited amount I've tested.
# Here, we transform the cookie from the server into something the client
#   will always send back to us, and embed the real server/path info in the
#   name of the name-value pair, since the cookie is uniquely identified by
#   the domain, path, and name.  Upon return from the client, we split the
#   name back into its original fields.
# One way to get around *some* of the all-cookies-all-the-time problem,
#   *sometimes*, may be possible to program with the following approach:
#   First, the target URL must be "encoded" (in proxy_encode()) in a way
#   that it resembles a path.  For example, the default "://" --> "/"
#   encoding does this.  Then, let the cookies go back to the client with
#   the target paths still intact.  This would only work when the cookie
#   domain is the default, i.e. the source host.  Check other possibilities
#   carefully, too, but I think you could get it to work somehow.
# Question-- is the port supposed to be used in the domain field?
#   Everything here assumes not, which is conceivably a security risk.

# Transform the cookie into something the client will send back through
#   the script, but still has all the needed info.  Returns undef if
#   cookie is invalid (e.g. comes from the wrong host).
# A cookie is uniquely identified by the domain, path, and name, so this 
#   transformation embeds the path and domain info into the "name".
sub cookie_to_client {
    local($cookie, $source_path, $source_server)= @_ ;
    local($name, $value, $expires_clause, $path, $domain, $secure_clause) ;
    local($new_name, $new_cookie) ;

    ($name,$value)= $cookie=~ /^\s*([^=;,\s]*)=?([^;,\s]*)/ ;
    ($expires_clause)= $cookie=~ /;\s*(expires\s*=[^;]*)/i ;
    ($path)= $cookie=~ /;\s*path\s*=\s*([^;,\s]*)/i ;      # clash w/ ;-params?
    ($domain)= $cookie=~ /;\s*domain\s*=\s*([^;,\s]*)/i ;
    ($secure_clause)= $cookie=~ /;(\s*secure\b)/i ;

    # Path defaults to path of URL that sent the cookie
    $path= $source_path if $path eq '' ;

    # Domain must be checked for validity: defaults to the server that sent
    #   the cookie; otherwise, must match end of that server name, and must
    #   contain at least two dots if in one of these seven top-level domains, 
    #   three dots otherwise.
    if ($domain eq '') {
	$domain= $source_server ;
    } else {
	return(undef) if $source_server!~ /$domain$/ ;
	return(undef) unless 
	    ( ( ($domain=~ tr/././) >= 3 ) ||
	      ( ($domain=~ tr/././) >= 2 && 
	        $domain=~ /\.(com|edu|net|org|gov|mil|int)$/ )
	    ) ;
    }

    # This is hereby the transformed format: name is $name;$path;$domain
    #   with %, ;, and = URL-encoded.  (The three values won't already
    #   have semicolons in them.)
    $new_name= join(';', $name, $path, $domain) ;
    $new_name=~ s/%/%25/g ;    # must be first
    $new_name=~ s/;/%3b/g ;
    $new_name=~ s/=/%3d/g ;

    # The new domain is this proxy server, which is the default if it is not 
    #   specified.
    $new_cookie= join('; ', grep(length,
				 $new_name . '=' . $value, 
				 $expires_clause, 
				 'path=' . $ENV{'SCRIPT_NAME'} . '/',
				 $secure_clause 
		     )) ;

    return($new_cookie) ;

}


# Unwrap the cookies from the client that were transformed earlier by the
#   routine cookie_to_client().  Returns composite cookie with all cookies
#   that match the domain and path, and no others!
# Currently, the name field contains the path and domain info-- see
#   cookie_to_client() for description of format.
sub cookie_to_server {
    local($cookie, $target_path, $target_server)= @_ ;
    local(@cookies)= split(/\s*;\s*/, $cookie) ;
    local($name, $value, $path, $domain, @matches, %pathlen) ;
    foreach (@cookies) {
	($name,$value)= split(/=/, $_, 2) ;
        $name=~ s/%3d/=/g ;
	$name=~ s/%3b/;/g ;
	$name=~ s/%25/%/g ;
	($name,$path,$domain)= split(/;/, $name) ;
	push(@matches, $name.'='.$value), 
	    $pathlen{$matches[$#matches]}= length($path)
		if ($target_server=~ /$domain$/i  && $target_path=~ /^$path/) ;
    }

    # More specific path mappings (i.e. longer paths) should be sent first.
    # If %pathlen is declared with my(), this generates a warning in Perl 5. :(
    sub reversepathlength { length($pathlen{$b}) <=> length($pathlen{$a}) }
    return join('; ', sort reversepathlength @matches) ;
}



#--------------------------------------------------------------------------
#    Utility routines
#--------------------------------------------------------------------------


# Escape any &"<> chars to &xxx; and return resulting strings.
sub HTMLescape {
    local($s)= @_ ;
    $s=~ s/&/&amp;/g ;      # must be before all others
    $s=~ s/"/&quot;/g ;
    $s=~ s/</&lt;/g ;
    $s=~ s/>/&gt;/g ;
    return $s ;
}


# Unescape any &xxx; codes back to &"<> and return resulting string.
# Simplified version here; only includes &"<>
sub HTMLunescape {
    local($s)= @_ ;
    $s=~ s/&quot;/"/g ;
    $s=~ s/&lt;/</g ;
    $s=~ s/&gt;/>/g ;
    $s=~ s/&amp;/&/g ;      # must be after all others
    return $s ;
}
    

#--------------------------------------------------------------------------
#    Output routines
#--------------------------------------------------------------------------

# Alert the user to an unsupported URL, with this intermediate page
sub unsupported_warning {
    print <<EOF ;
HTTP/1.0 200 OK
Content-type: text/html

<html>
<head><title>WARNING: Invalid proxy used!</title></head>
<body>
<h1>WARNING: Invalid proxy used!</h1>
<h3>This proxy only supports SSL ("https").  Any other URL needs to be
directed through the main proxy.</h3>
<p>
<hr>
<i>Based on <a href="http://www.jmarshall.com/tools/cgiproxy/">CGIProxy $PROXY_VERSION</a></i>
<p>
</body>
</html>
EOF

    exit ;
}


# Return "403 Forbidden" message, with explanatory text
sub nontextdie {
    print <<EOF ;
HTTP/1.0 403 Forbidden
Content-type: text/html

<html>
<head><title>Proxy cannot forward non-text files</title></head>
<body>
<h1>Proxy cannot forward non-text files</h1>
<p>Due to bandwidth limitations, the owner of this particular proxy is
forwarding only text files.  For best results, turn off automatic image 
loading if your browser lets you.
<p>If you need access to images or other binary data, route your browser 
through another proxy (or install one yourself--
<a href="http://www.jmarshall.com/tools/cgiproxy/">it's easy</a>).
<p>
<hr>
<i>Based on <a href="http://www.jmarshall.com/tools/cgiproxy/">CGIProxy $PROXY_VERSION</a></i>
<p>
</body>
</html>
EOF

    exit ;
}


# When an image should be blanked, returns either a transparent 1x1 GIF or
#   a 404 result.  Yes, that's an inlined 43-byte GIF.
sub skip_image {
    if ($RETURN_EMPTY_GIF) {
	print <<EOF ;
HTTP/1.0 200 OK
Content-Type: image/gif
Content-Length: 43

GIF89a\x01\0\x01\0\x80\0\0\0\0\0\xff\xff\xff\x21\xf9\x04\x01\0\0\0\0\x2c\0\0\0\0\x01\0\x01\0\x40\x02\x02\x44\x01\0\x3b
EOF
    } else {
	print "HTTP/1.0 404 Not Found\n\n" ;
    }

    exit ;
}


# Die, outputting HTML error page
sub HTMLdie {
    local($msg)= @_ ;
    print <<EOF ;
HTTP/1.0 200 OK
Content-Type: text/html

<html>
<head><title>CGI Proxy Error</title></head>
<body>
<h1>CGI Proxy Error</h1>
<h3>$msg</h3>
<p>
<hr>
<i>Based on <a href="http://www.jmarshall.com/tools/cgiproxy/">CGIProxy $PROXY_VERSION</a></i>
<p>
</body>
</html>
EOF

    exit ;
}
